/*
 * linux/arch/sh/kernel/irq_intc2.c
 *
 * Copyright (C) 2001 David J. Mckay (david.mckay@st.com)
 *
 * May be copied or modified under the terms of the GNU General Public
 * License.  See linux/COPYING for more information.                            
 *
 * Interrupt handling for INTC2-based IRQ.
 *
 * These are the "new Hitachi style" interrupts, as present on the 
 * Hitachi 7751 and the STM ST40 STB1.
 *
 * This module has been extended to support the SH7760
 *
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/irq.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/machvec.h>


struct sh7760_intc2_data {
	unsigned int addr;	/* Address of Interrupt Priority Register */
	int mask; /*Mask to apply */
	int mask_offset;	/* Mask register offset */
};


static struct sh7760_intc2_data intc2_data[NR_INTC2_IRQS];

static void sh7760_enable_intc2_irq(unsigned int irq);
static void sh7760_disable_intc2_irq(unsigned int irq);

/* shutdown is same as "disable" */
#define sh7760_shutdown_intc2_irq sh7760_disable_intc2_irq

static void sh7760_mask_and_ack_intc2(unsigned int);
static void sh7760_end_intc2_irq(unsigned int irq);

static unsigned int sh7760_startup_intc2_irq(unsigned int irq)
{ 
	sh7760_enable_intc2_irq(irq);
	return 0; /* never anything pending */
}

static struct hw_interrupt_type intc2_irq_type = {
	"INTC2-IRQ",
	sh7760_startup_intc2_irq,
	sh7760_shutdown_intc2_irq,
	sh7760_enable_intc2_irq,
	sh7760_disable_intc2_irq,
	sh7760_mask_and_ack_intc2,
	sh7760_end_intc2_irq
};

static void sh7760_disable_intc2_irq(unsigned int irq)
{
	unsigned addr;
	int offset=irq-INTC2_FIRST_IRQ;
	unsigned val,flags;

	// Sanity check
	if(offset<0 || offset>=NR_INTC2_IRQS) return;

	addr=intc2_data[offset].addr + intc2_data[offset].mask_offset;

	save_and_cli(flags);
	val=ctrl_inl(addr);
	val|=intc2_data[offset].mask;
	ctrl_outl(val,addr);

	restore_flags(flags);
}

static void sh7760_enable_intc2_irq(unsigned int irq)
{
	unsigned addr;
	int offset=irq-INTC2_FIRST_IRQ;

	// Sanity check
	if(offset<0 || offset>=NR_INTC2_IRQS) return;

	addr = intc2_data[offset].addr+ intc2_data[offset].mask_offset + 0x20;
	ctrl_outl(intc2_data[offset].mask, addr);

}

static void sh7760_mask_and_ack_intc2(unsigned int irq)
{
	sh7760_disable_intc2_irq(irq);
}

static void sh7760_end_intc2_irq(unsigned int irq)
{
	if (!(irq_desc[irq].status & (IRQ_DISABLED|IRQ_INPROGRESS)))
		sh7760_enable_intc2_irq(irq);
}

void sh7760_make_intc2_irq(unsigned int irq, unsigned int addr, 
		    unsigned int group,int pos, int priority, int mask_offset,
			int pri_offset)
{
	int offset=irq-INTC2_FIRST_IRQ;
	unsigned flags,val;

	if(offset<0 || offset>=NR_INTC2_IRQS) {
		return;
	}

	disable_irq_nosync(irq);
	/* Fill the the data we need */
	intc2_data[offset].addr=addr;
	intc2_data[offset].mask=1<<pos;
	intc2_data[offset].mask_offset = mask_offset;
		
	/* Set the priority level */
	save_and_cli(flags);
	val=ctrl_inl(addr+pri_offset);
	val|=(priority)<< (group*4);
	ctrl_outl(val,addr+pri_offset);
	restore_flags(flags);

	irq_desc[irq].handler=&intc2_irq_type;

	sh7760_disable_intc2_irq(irq);
}



