#include <linux/config.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/ioctl.h>
#include <linux/fs.h>
#include <linux/init.h>

#include <asm/uaccess.h>

#include "fctlio.h"

/*
 * Strata NOR Macros
 */
#define NOR_READ_ARRAY                0x00FF00FF
#define NOR_ERASE_SETUP               0x00200020
#define NOR_ERASE_CONFIRM             0x00D000D0
#define NOR_CMD_CLEAR_LOCK            0x00600060
#define NOR_CMD_WRITE                 0x00400040
#define NOR_PGM_SETUP                 CMD_WRITE
#define NOR_STATUS_READ               0x00700070
#define NOR_STATUS_CLEAR              0x00500050
#define NOR_CMD_READ_QUERY            0x00980098
#define NOR_CMD_WR_BUFF               0x00e800e8
#define NOR_CMD_CONFIRM               0x00D000D0

#define NOR_STATUS_PROTECT            0x00020002
#define NOR_STATUS_VOLTAGE_ERR        0x00080008
#define NOR_STATUS_PGM_ERR            0x00100010
#define NOR_STATUS_ERASE_ERR          0x00200020
#define NOR_STATUS_BUSY               0x00800080

#define NOR_POLLMAX                   (1<<25)
#define NOR_WR_BUFF_SIZE              32 //16
#define NOR_NUM_FLASH_BLOCKS          (128)
#define NOR_FLASH_BLOCK_BASE          ((unsigned long *) 0x040000)
#define NOR_FLASH_BLOCK_SIZE          ((unsigned long) 0x040000)

#define NOR_MAIN_BLOCK_SIZE           (0x40000)
#define NOR_MAIN_BLOCK_WORDS          (MAIN_BLOCK_SIZE / 2)
#define NOR_MAIN_BLOCK_DWORDS         (MAIN_BLOCK_WORDS / 2)
#define NOR_FLASH_MAX                 (MAIN_BLOCK_SIZE * NUM_FLASH_BLOCKS)
#define NOR_FLASH_BUS_WIDTH           (sizeof(unsigned long))

#define NAP_FACTORY_SIZE_ADDR         0xA007FFFC
#define NAP_FACTORY_SIZE()            (*(unsigned short *)NAP_FACTORY_SIZE_ADDR)
#define NAP_FACTORY_BARKER_ADDR       (NAP_FACTORY_SIZE_ADDR + 2)
#define NAP_FACTORY_BARKER            0x434A
#define NAP_FACTORY_EXIST()           (NAP_FACTORY_BARKER == *(unsigned short *)NAP_FACTORY_BARKER_ADDR)
#define NAP_FACTORY_MODEL_OFFSET      10
#define NAP_FACTORY_MODEL_MAXSIZE     32
#define NAP_FACTORY_MODEL_ADDR        (NAP_FACTORY_SIZE_ADDR - NAP_FACTORY_SIZE() + NAP_FACTORY_MODEL_OFFSET)
#define NAP_DEFAULT_MODEL             "NAP"

#define CAST(x) ((unsigned long)x)

static int flash_init(void);
static int flash_init_nand(void);
static int flash_init_strata(void);
static int flash_read_page(unsigned char* buffer, unsigned int address);
static int flash_read_page_nand(unsigned char* buffer, unsigned int address);
static int flash_read_page_strata(unsigned char* buffer, unsigned int address);
static int flash_read_oob_nand(unsigned char* buffer, unsigned int address);
static int flash_read_oob(unsigned char* buffer, unsigned int address);
static int flash_write_page(unsigned char* buffer, unsigned int address);
static int flash_write_page_nand(unsigned char* buffer, unsigned int address);
static int flash_write_page_strata(unsigned char* buffer, unsigned int address);
static int flash_erase_block(unsigned int address);
static int flash_erase_block_nand(unsigned int address);
static int flash_erase_block_strata(unsigned int address);
static void flash_get_id(unsigned char *manufacturer, unsigned char *device);
static void flash_get_id_nand(unsigned char *manufacturer, unsigned char *device);
static void flash_get_id_strata(unsigned char *manufacturer, unsigned char *device);
static void flash_read_model(char *user_data);

#define FCTLIO_MAJOR 127
#define NAND_PAGE_SIZE (528)

static ssize_t fctlio_read(struct file *file, char *buf,
                         size_t count, loff_t *off);
static ssize_t fctlio_write(struct file *file, const char *buf,
                         size_t count, loff_t *off);
static int fctlio_ioctl(struct inode *inode, struct file *file,
                     unsigned int op, unsigned long arg);
static int fctlio_open(struct inode *inode, struct file *file);
static int fctlio_close(struct inode *inode, struct file *file);

static struct file_operations fctlio_fops = {
  read:    fctlio_read,   /* read    */
  write:   fctlio_write,  /* write   */
  ioctl:   fctlio_ioctl,  /* ioctl   */
  open:    fctlio_open,   /* open    */
  release: fctlio_close,  /* release */
};

static ssize_t fctlio_read(struct file *file, char *buf, size_t count, loff_t
*off) {
        return 0;
}

static ssize_t fctlio_write(struct file *file, const char *buf, size_t count,
loff_t *off) {
        return 0;
}

static int fctlio_ioctl(struct inode *inode, struct file *file,
                        unsigned int op, unsigned long arg)
{
    nand_ctl user_arg;
  
    switch (op)
    {
        case FCTL_READ_CMD:
        {
	    //printk("NAND flash READ\n");
	    copy_from_user(&user_arg, (char*) arg, sizeof(nand_ctl));
	    if (!flash_read_page(user_arg.data, user_arg.address))
	    {
	        return -EINVAL;
	    }
	    copy_to_user((char*) arg, &user_arg, sizeof(nand_ctl));
	}
	break;
          
        case FCTL_READ_OOB_CMD:
        {
            //printk("NAND flash READ OOB\n");
	    copy_from_user(&user_arg, (char*) arg, sizeof(nand_ctl));
	    if (!flash_read_oob(user_arg.data, user_arg.address))
	    {
	        return -EINVAL;
	    }
	    copy_to_user((char*) arg, &user_arg, sizeof(nand_ctl));
	}
	break;

        case FCTL_WRITE_CMD:
        {
            //printk("NAND flash WRITE\n");
	    copy_from_user(&user_arg, (char*) arg, sizeof(nand_ctl));
	    if (!flash_write_page(user_arg.data, user_arg.address))
	    {
	        return -EINVAL;
	    }
	}
	break;
     
        case FCTL_ERASE_CMD:
        {
            //printk("NAND flash ERASE\n");
	    if (!flash_erase_block(arg))
	    {
	        return -EINVAL;
	    }
	}
	break;

        case FCTL_GET_ID_CMD:
        {
	    char device;
	    char manufacturer;
	    unsigned int value;

	    flash_get_id(&manufacturer, &device);
	    value = (device & 0xffff) | ((manufacturer << 16) & 0xffff0000);
	    copy_to_user((char*) arg, &value, sizeof(unsigned int));
	}
	break;

        case FCTL_GET_MODEL_CMD:
	{
            flash_read_model((char *)arg);
	}
	break;

        default:
	    return -EINVAL;
	    break;
    }

    return 0;
}

static int fctlio_open(struct inode *inode, struct file *file)
{
  return 0;
}

static int fctlio_close(struct inode *inode, struct file *file)
{
  return 0;
}

static int __init fctlio_init(void)
{
  printk("FCTL IO driver v0.1\r\n");

  if (register_chrdev(FCTLIO_MAJOR, "FCTLIO", &fctlio_fops)) {
    printk("unable to get major %d for FCTLIO_MAJOR\n", FCTLIO_MAJOR);
  }

  if (!flash_init())
  {
    printk("unable to initialze the nand flash device\n");

  }
  
  return 0;
}

void __exit fctlio_exit(void)
{
}

module_init(fctlio_init);
module_exit(fctlio_exit);

#include "sh4gpio.h"
#include "flctl.h"
#include "nandpriv.h"
#include "nandcmds.h"

#define itc_printf   printk

#define OOB_SIZE    (16)

static int flash_device_error;

/* Gets status register */
static unsigned char get_status_register(void)
{
    int timeout;
    unsigned char data;
    
    FLCMNCR = CE0 | ACM_COMMAND | TYPESEL; // Enable chip, NAND
    FLCMDCR = DOCMD1 | SCTCNT(1) | FLASH_READ; // Do command 1, read 1 sector
    FLCMCDR = CMD_0(READ_STATUS); // read status register
    FLDTCNTR = DTCNT(1); // 1 byte to transfer

    // Start transaction
    FLTRCR = TRSTART;
    timeout = 0;
    while (!(FLTRCR & TREND))
    {
        if ((timeout++) > TIMEOUT_VALUE)
        {
            FLCMNCR = 0;
            printk("Status timeout\n");
            return 0x01; // fail
        }
        else
        {
            udelay(10);
        }
    }
    FLTRCR = 0; // clear transfer
    
    //itc_printf("FLDATAR = 0x%x, FLBSYCNT = 0x%x\r\n", FLDATAR, FLBSYCNT);
    
    data = DATAR_0(FLDATAR);
    FLCMNCR = 0; // disable chip
    //itc_printf("Status = %b\r\n", data);
    
    return data;
}

/* Sets the flash pointer to either 0-255, 256-511, or 512-527
 * Offset has to be within that range
 */
static int move_flash_pointer(int offset)
{
    unsigned char command;
    int timeout;

    if (offset < 0)
    {
        return FALSE;
    }
    else if (offset < 256)
    {
        command = READ_MODE_1;
    }
    else if (offset < 512)
    {
        command = READ_MODE_2;
    }
    else 
    {
        command = READ_MODE_3;
    }

    FLCMNCR = CE0 | ACM_COMMAND | TYPESEL; // enable NAND controller
    FLCMDCR = DOCMD1 | SCTCNT(1) | FLASH_READ; // do command 1, read 1 sector
    FLCMCDR = CMD_0(command);
    FLDTCNTR = DTCNT(0); // we're not going to read anything, just issue the command.

    // start flash cycle
    FLTRCR = TRSTART;
    timeout = 0;
    while (!(FLTRCR & TREND))
    {
        if ((timeout++) > TIMEOUT_VALUE)
        {
            flash_device_error = FLASH_ERROR_TIMEOUT;
            printk("move flash pointer timeout\n");
            FLCMNCR = 0;
            return FALSE;
        }
        else
        {
            udelay(10);
        }
    }
    FLTRCR = 0; // clear transfer reg
    FLCMNCR = 0; // disable controller.

    return TRUE;
}

/* Gets ID information (manufacturer, device) */
void flash_get_id_nand(unsigned char *manufacturer, unsigned char *device)
{
    int timeout;
    
    FLCMNCR = CE0 | ACM_COMMAND | TYPESEL; // Enable chip, NAND
    FLCMDCR = ADRMD | DOCMD1 | SCTCNT(1) | FLASH_READ | DOADR | ADRCNT_1_BYTE; // do command 1, read 1 sector, address as is
    FLCMCDR = CMD_0(READ_ID); // read ID command
    FLDTCNTR = DTCNT(2); // read ID is 2 bytes (Samsung lets you read 4 - the next byte is 0xA5 ("Don't care") and 0xC0 ("support 
                         // multi-plane"). Sandisk returns similar.
    FLADR = ADR_0(0x00); // read address zero

    // Start transaction
    FLTRCR = TRSTART; 
    timeout = 0;
    while (!(FLTRCR & TREND))
    {
        if ((timeout++) > TIMEOUT_VALUE)
        {
            flash_device_error = FLASH_ERROR_TIMEOUT;
            FLCMNCR = 0;
            return;
        }
        else
        {
            udelay(1);
        }
    }
    FLTRCR = 0;
    *manufacturer = DATAR_0(FLDATAR);
    *device = DATAR_1(FLDATAR);
    //itc_printf("FLDATAR = 0x%x\r\n", FLDATAR);

    FLCMNCR = 0; // disable chip

#if 0
    // ID2
    FLCMNCR = CE0 | ACM_COMMAND | TYPESEL; // Enable chip, NAND
    FLCMDCR = ADRMD | DOCMD1 | SCTCNT(1) | FLASH_READ | DOADR | ADRCNT_1_BYTE; // do command 1, read 1 sector, address as is
    FLCMCDR = CMD_0(READ_ID_2); // read ID command
    FLDTCNTR = DTCNT(4); // read ID is 2 bytes
    FLADR = ADR_0(0x00); // read address zero
    FLDATAR = 0;
    itc_printf("FLDATAR = 0x%x\r\n", FLDATAR);

    // Start transaction
    FLTRCR = TRSTART; 
    timeout = 0;
    while (!(FLTRCR & TREND))
    {
        if ((timeout++) > TIMEOUT_VALUE)
        {
            flash_device_error = FLASH_ERROR_TIMEOUT;
            return;
        }
        else
        {
            udelay(1);
        }
    }
    FLTRCR = 0;
    itc_printf("FLDATAR = 0x%x\r\n", FLDATAR);
    
    FLCMNCR = 0; // disable chip
#endif
}

void flash_get_id_strata(unsigned char *manufacturer, unsigned char *device)
{
   /*
    * change to get Flash ID mode
    */
   NOR_ACCESS(NOR_BASE) = NOR_DEV_MAN_READ;
   *manufacturer = (unsigned char)NOR_ACCESS(NOR_BASE);
   *device = (unsigned char)NOR_ACCESS(NOR_BASE + 4);

   /*
    * Back to read mode
    */
   NOR_ACCESS(NOR_BASE) = NOR_READ_ARRAY;
}

/* Gets ID information (manufacturer, device) */
void flash_get_id(unsigned char *manufacturer, unsigned char *device)
{
    if (BOOT_FLASH_IS_32BIT())
      flash_get_id_strata(manufacturer, device);
    else
      flash_get_id_nand(manufacturer, device);
}
  
/* Writes a page of data to flash. Buffer is of NAND_PAGE_SIZE bytes.
 * Returns TRUE if flash was successful, FALSE otherwise
 * address is page address
 */
int flash_write_page_nand(unsigned char* buffer, unsigned int address)
{
    unsigned int left_to_transfer;
    int index;
    int timeout;
    unsigned char status;
    
    //printk("%s - buffer 0x%X, addr 0x%X\n", __FUNCTION__, buffer, address);
    if (!move_flash_pointer(0))
    {
        itc_printf("Error setting flash pointer for write - %d\r\n", flash_device_error);
        return FALSE;
    }
    
    // Setup transaction
    // Clear FLINTDMACR
    FLINTDMACR |= AC1CLR | AC0CLR; // clear FIFOs
    FLCMNCR = CE0 | ACM_COMMAND | TYPESEL | FCKSEL | NANDWF; // Enable chip, select command, select NAND, divide clock by 2, 
                                                            // address and data writes divide by 2
    FLCMDCR = ADRMD | CDSRC | FLASH_WRITE | DOADR | ADRCNT_4_BYTE | DOCMD1 | DOCMD2 | SCTCNT(1);
                                                            // Copy address register exactly to flash, select FIFO
                                                            // Write to flash, do address stage, 4 byte addresses
                                                            // do commnand 1, do command 2, and 1 sector
    FLCMCDR = CMD_0(SERIAL_DATA_INPUT) | CMD_1(AUTO_PROGRAM); // First command - input into buffer, second command - program
    FLDTCNTR = DTCNT(NAND_PAGE_SIZE); // We're going to transfer that many bytes.
    FLADR = ADR_0(0) | ADR_1((address & 0x0000FF) >> 0) | ADR_2((address & 0x00FF00) >> 8) | ADR_3((address & 0x030000) >> 16);
    
    left_to_transfer = NAND_PAGE_SIZE;
    // Reset status
    FLINTDMACR = 0;

    // Prefill the data fifo (prevent underruns)
    for (index = 0; index < 128; index += 4)
    {
        FLDTFIFO = DATAW_0(buffer[index]) | DATAW_1(buffer[index + 1]) | DATAW_2(buffer[index + 2]) | DATAW_3(buffer[index + 3]);
        left_to_transfer -= 4;
    }
    
    //printk("Transaction setup OK. FLCMNCR 0x%X, FLCMDCR 0x%X, FLCMCDR 0x%X, FLDTCNTR 0x%X, FLADR 0x%X\n", FLCMNCR, FLCMDCR, FLCMCDR, FLDTCNTR, FLADR);
    // Start transaction
    FLTRCR = TRSTART;

    // transfer rest
    while (left_to_transfer)
    {
        if (DTFLW(FLDTCNTR) > 0) // prevent overrun of buffer
        {
            FLDTFIFO = DATAW_0(buffer[index]) | DATAW_1(buffer[index + 1]) | DATAW_2(buffer[index + 2]) | DATAW_3(buffer[index + 3]);
            index += 4;
            left_to_transfer -= 4;
        }
    }

    // Wait for it to complete
    udelay(200); // programming time (typ)
    timeout = 0;
    while (!(FLTRCR & TREND))
    {
        if ((timeout++) > TIMEOUT_VALUE)
        {
            flash_device_error = FLASH_ERROR_TIMEOUT;
            FLCMNCR = 0;
            return FALSE;
        }
        else
        {
            udelay(1);
        }
    }

    // Transfer complete
    FLTRCR = 0; // clear transfer control

    // Analyze FLINTDMACR register
    if (FLINTDMACR & DTOERB)
    {
        itc_printf("FLDTFIFO Overrun!\r\n");
    }
    if (FLINTDMACR & DTUERB)
    {
        itc_printf("FLDTFIFO Underrun!\r\n");
    }
    
    // Disable controller
    FLCMNCR = 0;

    // Get status
    status = get_status_register();

    // Analyze status
    if ((status & STATUS_WRITEPROTECT) == STATUS_PROTECT)
    {
        flash_device_error = FLASH_ERROR_WRITEP;
        return FALSE;
    }

    if ((status & STATUS_READYBUSY) == STATUS_BUSY) // ??? should never happen
    {
        flash_device_error = FLASH_ERROR_UNKNOWN;
        return FALSE;
    }

    if ((status & STATUS_PASSFAIL) == STATUS_PASS)
    {
        flash_device_error = FLASH_ERROR_NONE;
        return TRUE;
    }
    else
    {
        flash_device_error = FLASH_ERROR_HWFAIL;
        return FALSE;
    }

    /* At the end, soemthing went horribly wrong - should never
     * get here*/
    flash_device_error = FLASH_ERROR_UNKNOWN;
    return FALSE;
}

int flash_write_page_strata(unsigned char* buffer, unsigned int address)
{
     address |= 0xa0000000;
     nor_block_write_flash_page((unsigned long *)address, (unsigned long *)buffer);

    return TRUE;
}

int flash_write_page(unsigned char* buffer, unsigned int address)
{
    if (BOOT_FLASH_IS_32BIT())
      return flash_write_page_strata(buffer, address);
    
    return flash_write_page_nand(buffer, address);
}

/* Reads a page of data from flash. Buffer is of NAND_PAGE_SIZE bytes.
 * Returns TRUE if read was successful, FALSE otherwise
 * address is page address
 */
int flash_read_page_nand(unsigned char* buffer, unsigned int address)
{
    unsigned int left_to_transfer;
    int index;
    int timeout;
    unsigned int data;
    
    //printk("flash_read_page - address 0x%X, addr 0x%X\n", buffer, address);
    // Setup transaction
    // FIFO direction change, clear FLINTDMACR
    FLINTDMACR |= AC1CLR | AC0CLR;
    FLCMNCR = CE0 | ACM_COMMAND | TYPESEL | FCKSEL | NANDWF; // enable chip, select command mode, NAND, fck/2, nand clk/2
    FLCMDCR = ADRMD | CDSRC | FLASH_READ | DOADR | ADRCNT_4_BYTE | DOCMD1 | SCTCNT(1); // address as-is, use FIFO, read flash
                                                            // do address cycle, 4 byte addresses, do command 1 stage, 1 sector
    FLCMCDR = CMD_0(READ_MODE_1);
    //printk("NAND_PAGE_SIZE = %d\n", NAND_PAGE_SIZE);
    FLDTCNTR = DTCNT(NAND_PAGE_SIZE); // we're going to transfer NAND_PAGE_SIZE bytes
    FLADR = ADR_0(0) | ADR_1((address & 0x0000FF) >> 0) | ADR_2((address & 0x00FF00) >> 8) | ADR_3((address & 0x030000) >> 16);

    left_to_transfer = NAND_PAGE_SIZE;
    index = 0;

    // Reset status register
    FLINTDMACR = 0;

    //printk("Transaction setup OK. FLCMNCR 0x%X, FLCMDCR 0x%X, FLCMCDR 0x%X, FLDTCNTR 0x%X, FLADR 0x%X\n", FLCMNCR, FLCMDCR, FLCMCDR, FLDTCNTR, FLADR);
    // Start transaction
    FLTRCR = TRSTART;

    while (left_to_transfer)
    {
       if (DTFLW(FLDTCNTR) > 0)
       {
           data = FLDTFIFO;
           buffer[index + 0] = DATAR_0(data);
           buffer[index + 1] = DATAR_1(data);
           buffer[index + 2] = DATAR_2(data);
           buffer[index + 3] = DATAR_3(data);
           left_to_transfer -= 4;
           index += 4;
       }
    }
    
    timeout = 0;
    //printk("Transaction complete\n");
    // This really isn't needed...
    while (!(FLTRCR & TREND))
    {
        if ((timeout++) > TIMEOUT_VALUE)
        {
            flash_device_error = FLASH_ERROR_TIMEOUT;
            FLCMNCR = 0;
            printk("Timeout\n");
            return FALSE;
        }
        else
        {
            udelay(1);
        }
    }

    //printk("Wrapping up\n");
    // Transfer finished
    FLTRCR = 0;
    // Check for errors

    if (FLINTDMACR & DTOERB)
    {
        itc_printf("FLDTFIFO Overrun!\r\n");
    }
    if (FLINTDMACR & DTUERB)
    {
        itc_printf("FLDTFIFO Underrun!\r\n");
    }
    
    FLCMNCR = 0;

    flash_device_error = FLASH_ERROR_NONE;
    //printk("Done flash_read_page\n");
    return TRUE;
}

int flash_read_page_strata(unsigned char* buffer, unsigned int address)
{
  memcpy(buffer, (void *)address, 512 /*NAND_PAGE_SIZE*/);
  return TRUE;
}

int flash_read_page(unsigned char* buffer, unsigned int address)
{
    if (BOOT_FLASH_IS_32BIT())
      return flash_read_page_strata(buffer, address);

    return flash_read_page_nand(buffer, address);
}

/* Reads OOB data from flash. Buffer is of OOB_SIZE bytes.
 * Returns TRUE if read was successful, FALSE otherwise
 * address is page address
 */
int flash_read_oob_nand(unsigned char* buffer, unsigned int address)
{
    int index;
    int timeout;
    unsigned int data;
    
    //printk("%s - buffer 0x%X, addr 0x%X\n", __FUNCTION__, buffer, address);
    // Setup transaction
    // FIFO direction change, clear FLINTDMACR
    FLINTDMACR |= AC1CLR | AC0CLR;
    FLCMNCR = CE0 | ACM_COMMAND | TYPESEL | FCKSEL | NANDWF; // enable chip, select command mode, NAND, fck/2, nand clk/2
    FLCMDCR = ADRMD | CDSRC | FLASH_READ | DOADR | ADRCNT_4_BYTE | DOCMD1 | SCTCNT(1); // address as-is, use FIFO, read flash
                                                            // do address cycle, 4 byte addresses, do command 1 stage, 1 sector
    FLCMCDR = CMD_0(READ_MODE_3); // read OOB
    FLDTCNTR = DTCNT(OOB_SIZE); // we're going to transfer OOB bytes
    FLADR = ADR_0(0) | ADR_1((address & 0x0000FF) >> 0) | ADR_2((address & 0x00FF00) >> 8) | ADR_3((address & 0x030000) >> 16);

    index = 0;

    // Reset status register
    FLINTDMACR = 0;

    // Start transaction
    FLTRCR = TRSTART;

    //udelay(25); // OOB can fit in FIFO, so wait for transaction to complete first.
    timeout = 0;
    // This really isn't needed...
    while (!(FLTRCR & TREND))
    {
        if ((timeout++) > TIMEOUT_VALUE)
        {
            flash_device_error = FLASH_ERROR_TIMEOUT;
            FLCMNCR = 0;
            return FALSE;
        }
        else
        {
            udelay(1);
        }
    }

    // Transfer finished
    FLTRCR = 0;

    for (index = 0; index < OOB_SIZE; index += 4)
    {
        data = FLDTFIFO;
        buffer[index + 0] = DATAR_0(data);
        buffer[index + 1] = DATAR_1(data);
        buffer[index + 2] = DATAR_2(data);
        buffer[index + 3] = DATAR_3(data);
    }
    
    FLCMNCR = 0;

    flash_device_error = FLASH_ERROR_NONE;
    return TRUE;
}

int flash_read_oob(unsigned char* buffer, unsigned int address)
{
    if (BOOT_FLASH_IS_32BIT())
    {
       printk("Warning strata read OOB data not supported, address 0x%X\n", address);
       return TRUE;
    }

    return flash_read_oob_nand(buffer, address);
}

/* Erases a flash block. Returns TRUE if flash erased successfully,
 * FALSE otherwise
 * Address must be block-aligned.
 */
int flash_erase_block_nand(unsigned int address)
{
    int timeout;
    unsigned char status;
    unsigned int block;
    
    //printk("%s - addr 0x%X\n", __FUNCTION__, address);
    block = address & 0x0003FFE0; // Masks out block address (13 bits)

    // Setup transaction
    FLCMNCR = CE0 | TYPESEL; // Select NAND chip
    FLCMDCR = ADRMD | DOADR | ADRCNT_3_BYTE | DOCMD1 | DOCMD2 | SCTCNT(1); 
        // Use literal address, do address, 3 byte addresses
        // Do command 1, do command 2, 1 sector
    FLCMCDR = CMD_0(AUTO_BLOCK_ERASE) | CMD_1(AUTO_BLOCK_ERASE_CONFIRM); // command
    FLDTCNTR = DTCNT(0); // zero bytes transferred
    FLADR = ADR_0((block & 0x0000FF) >> 0) | ADR_1((block & 0x00FF00) >> 8) | ADR_2((block & 0xFF0000) >> 16);

    //printk("Transaction setup OK. FLCMNCR 0x%X, FLCMDCR 0x%X, FLCMCDR 0x%X, FLDTCNTR 0x%X, FLADR 0x%X\n", FLCMNCR, FLCMDCR, FLCMCDR, FLDTCNTR, FLADR);    // Start transaction
    FLTRCR = TRSTART;
    udelay(2000); // 2ms typ delay
    timeout = 0;
    while (!(FLTRCR & TREND))
    {
        // gives us a timeout of around 1 second
        if ((timeout++) > TIMEOUT_VALUE)
        {
            flash_device_error = FLASH_ERROR_TIMEOUT;
            FLCMNCR = 0;
            return FALSE;
        }
        else
        {
            udelay(500);
        }
    }
    FLTRCR = 0;
    FLCMNCR = 0;
    
    status = get_status_register();

    // Analyze status
    if ((status & STATUS_WRITEPROTECT) == STATUS_PROTECT)
    {
        flash_device_error = FLASH_ERROR_WRITEP;
        return FALSE;
    }

    if ((status & STATUS_READYBUSY) == STATUS_BUSY) // ??? should never happen
    {
        flash_device_error = FLASH_ERROR_UNKNOWN;
        return FALSE;
    }

    if ((status & STATUS_PASSFAIL) == STATUS_PASS)
    {
        flash_device_error = FLASH_ERROR_NONE;
        return TRUE;
    }
    else
    {
        flash_device_error = FLASH_ERROR_HWFAIL;
        return FALSE;
    }

    /* At the end, soemthing went horribly wrong - should never
     * get here*/
    flash_device_error = FLASH_ERROR_UNKNOWN;
    return FALSE;
}

static int
status_flash(volatile unsigned long *flash_addr)
{
   unsigned long flash_status = 0;
   int error = 0;

   flash_status = NOR_ACCESS(flash_addr);
   if(flash_status & NOR_STATUS_PGM_ERR)
   {
      printk("Strata FLASH_PROGRAMMING_ERROR\n");
      error = 1;
   }
   if(flash_status & NOR_STATUS_PROTECT)
   {
      printk("Strata FLASH_PROTECTED_ERROR\n");
      error = 1;
   }
   if(flash_status & NOR_STATUS_VOLTAGE_ERR)
   {
      printk("Strata FLASH_PROGRAMMING_ERROR\n");
      error = 1;
   }
   if(flash_status & NOR_STATUS_ERASE_ERR)
   {
      printk("Strata FLASH_ERASE_ERROR\n");
      error = 1;
   }

   if(error)
   {
      printk("flash_status: 0x%x\r\n", flash_status);
      NOR_ACCESS(flash_addr) = CAST(NOR_STATUS_CLEAR); /* kkw add */
   }

   return error;
}

int 
nor_wait_flash(volatile unsigned long *flash)
{
   int pollmax = NOR_POLLMAX;
   unsigned long flash_status;

   do
   {
      flash_status = NOR_ACCESS(flash);
   } while(--pollmax && (flash_status & CAST(NOR_STATUS_BUSY)) !=
                         CAST(NOR_STATUS_BUSY));

   if(pollmax)
   {
      return 1;
   }
   else
   {
      printk("Strata FLASH_DEAD_ERROR\n");
      return 0;
   }
} // end of nor_wait_flash

int nor_block_erase_flash(unsigned long *addr)
{
   volatile unsigned long *flash = addr;
   int error = 0;

#if 0
   NOR_ACCESS(flash) = CAST(NOR_STATUS_CLEAR);

   NOR_ACCESS(flash) = CAST(NOR_CMD_CLEAR_LOCK);
   NOR_ACCESS(flash) = CAST(NOR_CMD_CONFIRM);

   NOR_ACCESS(flash) = CAST(NOR_READ_ARRAY); /* kkw add */
#endif

   NOR_ACCESS(flash) = CAST(NOR_ERASE_SETUP);
   NOR_ACCESS(flash) = CAST(NOR_ERASE_CONFIRM);

   if(!nor_wait_flash(flash))
   {
      return 0;
   }

   error = status_flash(flash);
   if(error)
   {
      printk("bad erase: 0x%x\r\n", (unsigned long)flash);
   }

   NOR_ACCESS(flash) = CAST(NOR_READ_ARRAY);

   return !error;
} // end of nor_block_erase_flash

int
nor_block_write_flash_page(unsigned long *dest,
                         unsigned long *src)
{
   unsigned long i = 0;
   volatile u8 *flash = (u8 *)dest;
   u8 *data = (u8 *)src;
   int error = 0;
   int pollmax = NOR_POLLMAX;

   /* Debug */
   //printk("nor_block_write_file_sector 0x%x -> 0x%x\r\n", (unsigned long)src, (unsigned long)dest);

   while(i < 512)
   {
     //int cnt = NOR_WR_BUFF_SIZE / sizeof(unsigned short);
      int cnt = NOR_WR_BUFF_SIZE;

      do
      {
         NOR_ACCESS(flash) = CAST(NOR_CMD_WR_BUFF);
      } while(--pollmax && (NOR_ACCESS(flash) & CAST(NOR_STATUS_BUSY)) !=
                             CAST(NOR_STATUS_BUSY));

      if(!pollmax)
      {
         printk("Strata FLASH_DEAD_ERROR\n");
         return 0;
      }

#if 0
      NOR_ACCESS(flash) = CAST((NOR_WR_BUFF_SIZE / sizeof(unsigned short) - 1) |
                             ((NOR_WR_BUFF_SIZE / sizeof(unsigned short) - 1) <<
                              (NOR_FLASH_BUS_WIDTH * 4)));
#else
      NOR_ACCESS(flash) = CAST((NOR_WR_BUFF_SIZE - 1) |
                              ((NOR_WR_BUFF_SIZE - 1) <<
                               (NOR_FLASH_BUS_WIDTH * 4)));
#endif
      while( cnt-- )
      {
         NOR_ACCESS((flash + i)) = NOR_ACCESS(data);
         i += NOR_FLASH_BUS_WIDTH;
         data += NOR_FLASH_BUS_WIDTH;
      }
      NOR_ACCESS((flash + i-NOR_FLASH_BUS_WIDTH)) = CAST(NOR_CMD_CONFIRM);

      if(!nor_wait_flash((unsigned long *)flash))
      {
         return 0;
      }

      error = status_flash((unsigned long *)(flash + i-NOR_FLASH_BUS_WIDTH));
      if(error)
      {
         printk("bad addr: %x\r\n", (unsigned long)(flash + i));
         break;
      }
   }

   NOR_ACCESS(dest) = CAST(NOR_READ_ARRAY);
   return error;
}


int flash_erase_block_strata(unsigned int address)
{
    int i = 0;
    address = 0xa0000000 | (address & ~0x03);
    int block_size = 0x40000;

    NOR_ACCESS(address) = CAST(NOR_STATUS_CLEAR);

    NOR_ACCESS(address) = CAST(NOR_CMD_CLEAR_LOCK);
    NOR_ACCESS(address) = CAST(NOR_CMD_CONFIRM);

    NOR_ACCESS(address) = CAST(NOR_READ_ARRAY); /* kkw add */

    if (address >= 0xa7fc0000)
       block_size = 0x10000;

    /* erase check */
    while (i < block_size && 0xffffffff == ((unsigned long *)address)[i])
       ++i;
    
    if (i < block_size)
    {
       /* erase block */
       nor_block_erase_flash((unsigned long *)address);
    }

    return TRUE;
}

int flash_erase_block(unsigned int address)
{
    if (BOOT_FLASH_IS_32BIT())
      return flash_erase_block_strata(address);

    return flash_erase_block_nand(address);
}


static void flash_read_model(char *user_data)
{
    char model_name[NAP_FACTORY_MODEL_MAXSIZE+1];

    // NAP model
    if (NAP_FACTORY_EXIST() && (NAP_FACTORY_SIZE() > NAP_FACTORY_MODEL_OFFSET + NAP_FACTORY_MODEL_MAXSIZE))
    {
        // "JC" barker and factory data contains model name
	strncpy(model_name, (char *)NAP_FACTORY_MODEL_ADDR, NAP_FACTORY_MODEL_MAXSIZE);
    }
    else
    {
        strncpy(model_name, NAP_DEFAULT_MODEL, NAP_FACTORY_MODEL_MAXSIZE);
    }
    model_name[NAP_FACTORY_MODEL_MAXSIZE] = 0;
    copy_to_user(user_data, model_name, NAP_FACTORY_MODEL_MAXSIZE+1);
}

/* Initializes NAND flash */
static int flash_init_nand(void)
{
    int timeout;
    unsigned char status;
    
    // Select FLCTL controller mode (IPSELR)
    (*(volatile unsigned short *) (GPIO_BASE + IPSELR)) |= 0x0200; // Bit 9 set == FLCTL
    //printk("IPSELR = 0x%X\n", (*(volatile unsigned short *) (GPIO_BASE + IPSELR)));
    (*(volatile unsigned short *) (GPIO_BASE + PDCR  )) = 0; // Select Peripheral GPIO mode (data)
    //printk("PDCR   = 0x%X\n", (*(volatile unsigned short *) (GPIO_BASE + PDCR)));
    (*(volatile unsigned char  *) (GPIO_BASE + PDPUPR)) = 0; // disable pullups
    //printk("PDPUPR = 0x%X\n", (*(volatile unsigned char  *) (GPIO_BASE + PDPUPR)));
    (*(volatile unsigned short *) (GPIO_BASE + PKCR  )) = 0; // Select peripheral GPIO mode (control pins)
    //printk("PKCR   = 0x%X\n", (*(volatile unsigned short *) (GPIO_BASE + PKCR)));
    (*(volatile unsigned char  *) (GPIO_BASE + PKPUPR)) = 0; // disable pullups on control pins
    //printk("PKPIPR = 0x%X\n", (*(volatile unsigned char  *) (GPIO_BASE + PKPUPR)));
    // Set clocking
    // PLL = 200, CPU = 1/1 (200), Bus = 1/3 (66), Periph = 1/6 (33)
    // Leaves us with a choice of Fclks of 33MHz (1/1), 16MHz (1/2), 8MHz (1/4) or 4MHz (1/8)
    (*(volatile unsigned int   *) (0xFE0A0024)) = 1; // Lets start with 8.33MHz
    //printk("Clock = 0x%X\n", (*(volatile unsigned int *) (0xFE0A0024)));

    // Initialize flash chip (send reset command)
    FLCMNCR = CE0 | TYPESEL | ACM_COMMAND; // Select chip, NAND, command mode
    FLCMDCR = ADRMD | DOCMD1 | SCTCNT(1); // Execute first command only
    FLCMCDR = CMD_0(RESET); // Issue reset command
    FLDTCNTR = DTCNT(0); // No data to transfer

    FLTRCR = TRSTART; // Start transaction
    timeout = 0;
    udelay(500); // handle worst-case reset time
    while (!(FLTRCR & TREND))
    {
        if ((timeout++) < TIMEOUT_VALUE)
        {
            udelay(1);
        }
        else
        {
            flash_device_error = FLASH_ERROR_TIMEOUT;
            FLCMNCR = 0;
            return FALSE;
        }
    }
    FLTRCR = 0; // clear TREND & TRSTART
    FLCMNCR = 0; // clear CE0.

    if (!move_flash_pointer(0))
    {
        itc_printf("Error setting flash pointer\r\n");
        return FALSE;
    }

    status = get_status_register();
    
    return TRUE;
}

/* Initializes flash */
int flash_init_strata(void)
{
   /*
    * Just make sure flash is in read array mode
    */ 
   NOR_ACCESS(NOR_BASE) = NOR_READ_ARRAY;

   return TRUE;
}

/* Initializes flash */
int flash_init(void)
{
    if (BOOT_FLASH_IS_32BIT())
      return flash_init_strata();

    return flash_init_nand();
}
