/* 
 * linux/arch/sh/jci_n40/io.c
 *
 * Copyright (C) 2001  Ian da Silva, Jeremy Siegel
 * Based largely on io_se.c.
 *
 * I/O routine for JCI N40.
 *
 *
 * History:
 *   02/10/2006 cwellek Support cobra w/ strata flash
 */

#include <linux/kernel.h>
#include <linux/types.h>
#include <asm/io.h>
#include <asm/io_jci_n40.h>
#include <asm/addrspace.h>

#include <linux/pci.h>

// BSC
#define BSC_BCR2 0xff800004

// Test boot flash
#define BOOT_FLASH_IS_32BIT() (0xc000 == ((*(volatile unsigned short *)BSC_BCR2) & 0xc000))

/* Serial console port suppression */
#define SERIAL_CONSOLE_SUPP_ADDR 0xfe2e0600
#define SERIAL_CONSOLE_SUPP_FLAG 0x70707573
#define SERIAL_CONSOLE_SUPPRESS() (SERIAL_CONSOLE_SUPP_FLAG == *(unsigned long *)SERIAL_CONSOLE_SUPP_ADDR)

#define SMC_IOADDR		0xB0000300
#define DUART_BASE_A    (BOOT_FLASH_IS_32BIT() ? 0xB8000008 : 0xA4000008)

#define maybebadio(name,port) \
  printk("bad PC-like io %s for port 0x%lx at 0x%08x\n", \
	 #name, (port), (__u32) __builtin_return_address(0))

static inline void delay(void)
{
	ctrl_inw(0xa0000000);
}

/* Map the Ethernet addresses as if it is at 0x300 - 0x320 */
/* Map the DUART address as if it is at 0x3F8 */

static inline unsigned long __port2addr(unsigned long port)
{
     if (port >= 0x300 && port < 0x320)
     {
	  return SMC_IOADDR + (port - 0x300);
     }

     if (port >= 0x3F0 && port <= 0x3FF)
     {
	  return DUART_BASE_A + (port - 0x3F8);
     }
     
     maybebadio(__port2addr, port);
     return port;
}

unsigned char sh_jci_n40_inb(unsigned long port)
{
	if (port >= 0x300 && port < 0x320 && port & 0x01) {
		return (volatile unsigned char)(sh_jci_n40_inw(port -1) >> 8);
	}
	return *(volatile unsigned char *)__port2addr(port);
}

unsigned char sh_jci_n40_inb_p(unsigned long port)
{
	unsigned char v;

 	v = *(volatile unsigned char *)__port2addr(port);
        return v;
}

unsigned short sh_jci_n40_inw(unsigned long port)
{
    return *(volatile unsigned short *)__port2addr(port);
}

unsigned int sh_jci_n40_inl(unsigned long port)
{
    return *(volatile unsigned long *)__port2addr(port);
}

void sh_jci_n40_outb(unsigned char value, unsigned long port)
{
	if (port >= 0x300 && port < 0x320) {
        if (port & 0x01) {
            sh_jci_n40_outw(((unsigned short)value << 8), port -1);
            return;
        } else {
            sh_jci_n40_outw(((unsigned short)value), port);
            return;
        }
    }

	if ((port != 0x3f8) || !SERIAL_CONSOLE_SUPPRESS())
	*(volatile unsigned char *)__port2addr(port) = value;
}

void sh_jci_n40_outb_p(unsigned char value, unsigned long port)
{
	if ((port != 0x3f8) || !SERIAL_CONSOLE_SUPPRESS())
	*(volatile unsigned char *)__port2addr(port) = value;
}

void sh_jci_n40_outw(unsigned short value, unsigned long port)
{
	*(volatile unsigned short *)__port2addr(port) = value;
}

void sh_jci_n40_outl(unsigned int value, unsigned long port)
{
    //printk("%s: port = 0x%x, value = 0x%x translated addr - 0x%x\n", __FUNCTION__, port, value, __port2addr(port));
	*(volatile unsigned long *)__port2addr(port) = value;
}

void sh_jci_n40_insb(unsigned long port, void *addr, unsigned long count)
{
	unsigned char *p = addr;
	while (count--) *p++ = sh_jci_n40_inb(port);
}

void sh_jci_n40_insw(unsigned long port, void *addr, unsigned long count)
{
	unsigned short *p = (unsigned short*)addr;
	while (count--) *p++ = sh_jci_n40_inw(port);
}

void sh_jci_n40_insl(unsigned long port, void *addr, unsigned long count)
{
	unsigned long *p = (unsigned long*)addr;
	while (count--) *p++ = sh_jci_n40_inl(port);
}

void sh_jci_n40_outsb(unsigned long port, const void *addr, unsigned long count)
{
	unsigned char *p = (unsigned char*)addr;
	while (count--) sh_jci_n40_outb(*p++, port);
}

void sh_jci_n40_outsw(unsigned long port, const void *addr, unsigned long count)
{
	unsigned short *p = (unsigned short*)addr;
	while (count--) sh_jci_n40_outw(*p++, port);
}

void sh_jci_n40_outsl(unsigned long port, const void *addr, unsigned long count)
{
	unsigned long *p = (unsigned long*)addr;
	while (count--) sh_jci_n40_outl(*p++, port);
}

/* For read/write calls, just copy generic (pass-thru); PCIMBR is  */
/* already set up.  For a larger memory space, these would need to */
/* reset PCIMBR as needed on a per-call basis...                   */

unsigned char sh_jci_n40_readb(unsigned long addr)
{
	return *(volatile unsigned char*)addr;
}

unsigned short sh_jci_n40_readw(unsigned long addr)
{
	return *(volatile unsigned short*)addr;
}

unsigned int sh_jci_n40_readl(unsigned long addr)
{
	return *(volatile unsigned long*)addr;
}

void sh_jci_n40_writeb(unsigned char b, unsigned long addr)
{
    //if ((addr >= 0x300) && (addr <= 0x320))
    //{
    //    printk("%s called: addr 0x%x, data 0x%x\n", addr, b);
    //}
  if (addr != DUART_BASE_A || !SERIAL_CONSOLE_SUPPRESS())
	*(volatile unsigned char*)addr = b;
}

void sh_jci_n40_writew(unsigned short b, unsigned long addr)
{
    //if ((addr >= 0x300) && (addr <= 0x320))
    //{
    //    printk("%s called: addr 0x%x, data 0x%x\n", addr, b);
    //}
	*(volatile unsigned short*)addr = b;
}

void sh_jci_n40_writel(unsigned int b, unsigned long addr)
{
    //if ((addr >= 0x300) && (addr <= 0x320))
    //{
    //    printk("%s called: addr 0x%x, data 0x%x\n", addr, b);
    //}
        *(volatile unsigned long*)addr = b;
}


unsigned long
sh_jci_n40_isa_port2addr(unsigned long offset)
{
	return __port2addr(offset);
}
