/* NAND.H - Public NAND code defines/prototypes */
#ifndef NAND_H
#define NAND_H

/* This works for 512+16 flash only */
// define if using Sandisk MLC
#undef SANDISK

/* Flash IDs */
#define FLASH_TOSHIBA   (0x98)
#define FLASH_SAMSUNG   (0xEC)

/* ****** PHYSICAL NAND CONSTANTS ****** */
/* Size of data read */
#define DATA_SIZE   (512)
/* Size of OOB Area */
#define OOB_SIZE    (16)
/* Size of NAND page */
#define NAND_PAGE_SIZE   (DATA_SIZE + OOB_SIZE)
/* Size of NAND block, in pages */
#ifdef SANDISK
#define BLOCK_SIZE  (64) // for Sandisk MLC
#else
#define BLOCK_SIZE  (32)
#endif
/* Number of bits to shift to select a page */
#define PAGE_SHIFT  (9)
/* Number of bits to shift to select a block */
#ifdef SANDISK
#define BLOCK_SHIFT (15)
#else
#define BLOCK_SHIFT (14)
#endif
/* When doing multi-page programming, the number of blocks touched in one go */
#define INTERLEAVE_SHIFT    (2)
#define INTERLEAVE_FACTOR   (1 << INTERLEAVE_SHIFT)

/* Combo Blocks - For StingRay MULTI_BLOCK mode */
#define LARGE_BLOCK     (INTERLEAVE_FACTOR) // number of blocks per large block
#define LARGE_BLOCK_TO_ADDR(block) ((block << INTERLEAVE_SHIFT) << (BLOCK_SHIFT - PAGE_SHIFT))

// Takes a flash address and extracts just the block number out of it
#define ADDR_TO_LARGE_BLOCK(addr) ((addr) >> (BLOCK_SHIFT + INTERLEAVE_SHIFT))
#define LARGE_BLOCK_OFFSET(addr) ((addr) - (ADDR_TO_LARGE_BLOCK(addr) << (BLOCK_SHIFT + INTERLEAVE_SHIFT)))
#define LARGE_BLOCK_ISALIGNED(addr) (LARGE_BLOCK_OFFSET(addr) == 0)

/* For speed, scan only the first N blocks defined below */
#ifdef SANDISK
#define NUM_BLOCKS          (4096)
#else
#define NUM_BLOCKS          (8192)
#endif
#define NUM_LARGE_BLOCKS    (NUM_BLOCKS / LARGE_BLOCK)

/* Public API */

/* NOTE: All transfers are a full PAGE_SIZE bytes long. Short reads will be
 * extended to the encompassing page region. If the buffer isn't large enough,
 * a full page will be transferred, but not copied to the buffer
 *
 * Likewise, if the buffer is too short, the buffer will be padded with 0xFF's
 * to fill up the rest of the page.
 */

// Error codes
typedef enum 
{
    NoError     = 0, // No Error
    ECCFail     = 1, // ECC failed reading a page, uncorrectable
    HWFail      = 2, // Hardware failure
    Timeout     = 3, // Timeout
    WriteProtect = 4, // NAND is write protected
    Unknown     = 5, // Unknown problem
    NotInitialized = 6, // Flash not initialized
    OutOfRange  = 7, // Flash address out of range
    BadBlock    = 8  // Bad block (shouldn't ever happen, but may be returned internally)
} FlashError_t;

/* Init - call to initialize everything */
/* Returns 1 if successful, 0 if failed */
int nand_init_flash(void);

/* These functions only use Stingray sized Large Blocks */

/* Read from flash, given a block address */
/* Buffer is length bytes long, and only whole blocks can be read
 * (block-oriented). A higher layer will have to translate this down to
 * characters. Parial block reads are acceptable, though.
 */
int nand_read_block(unsigned int block_number, unsigned char *buffer, unsigned int length);

/* Write to flash, given a block address, and data */
/* Partial writes are possible. buffer is length bytes long */
int nand_write_block(unsigned int block_number, unsigned char *buffer, unsigned int length);

/* Erases a block */
int nand_erase_block(unsigned int block_number, unsigned int length);

#endif // NAND_H
