/* NANDPRIV.H - Private NAND flash code */

#ifndef NANDPRIV_H
#define NANDPRIV_H

/* Common */
#ifndef TRUE
#define TRUE            (1)
#endif

#ifndef FALSE
#define FALSE           (0)
#endif

#define NB_OF(x)    (sizeof(x) / sizeof(x[0]))

#define TIMEOUT_VALUE   (2000)

/* ***** ECC ***** */

/* Defines for ECC */
#define ECC_DATA_SIZE   (5)
#define ECC_OOB_SIZE    (3)

/* Error codes for ecc_correct() */
#define ECC_OK          (0)
#define ECC_CORRECTED   (1)
#define ECC_PARITYERROR (2)
#define ECC_UNCORRECTED (3)

/* Creates ECC */
void ecc_encode(unsigned char *data, unsigned char *oob_area);

/* Verifies ECC, corrects if necessary */
int ecc_correct(unsigned char *data, unsigned char *oob_area);

/* ***** DEVICE.C ***** */
extern int flash_device_error; /* Flash device error */
#define FLASH_ERROR_NONE    (0) /* No Error */
#define FLASH_ERROR_ECCFAIL (1) /* ECC check failed */
#define FLASH_ERROR_HWFAIL  (2) /* Hardware reported program error */
#define FLASH_ERROR_TIMEOUT (3) /* timed out waiting for flash */
#define FLASH_ERROR_WRITEP  (4) /* Write protected */
#define FLASH_ERROR_UNKNOWN (5) /* Wierd error occurred */
#define FLASH_ERROR_NOTINIT (6) /* Flash was not initialized */

/* Gets manufacturer name */
const char* flash_get_manufacturer_name(unsigned char manufacturer_code);

/* Gets ID information (manufacturer, device) */
void flash_get_id(unsigned char *manufacturer, unsigned char *device);

/* Gets device name */
const char* flash_get_device_name(unsigned char device_code);

/* Some flash devices have bad block in different location 
 * Almost all 528-byte page devices use offset 517 (byte 518), though
 * Return TRUE if block is bad. Return FALSE otherwise.
 */
int flash_is_bad_block(unsigned char* oob);

/* This is used to mark a block as bad in flash - according to the
 * datasheets, if a program or erase operation failed, the block should
 * be marked as bad. This function will take the OOB data and mark it in 
 * such a way that when the block is scanned, it will be marked as bad
 */
void flash_mark_bad_block(unsigned char* oob);

/* Writes a page of data to flash. Buffer is of PAGE_SIZE bytes.
 * Returns TRUE if flash was successful, FALSE otherwise
 * address is page address
 * NOTE - page must be erased, or if doing partial page programming
 * the unprogrammed bytes must be 0xFF.
 */
int flash_write_page(unsigned char* buffer, unsigned int address);

/* Reads a page of data from flash. Buffer is of PAGE_SIZE bytes.
 * Returns TRUE if read was successful, FALSE otherwise
 * address is page address
 */
int flash_read_page(unsigned char* buffer, unsigned int address);

/* Reads OOB data from flash. Buffer is of OOB_SIZE bytes.
 * Returns TRUE if read was successful, FALSE otherwise
 * address is page address
 */
int flash_read_oob(unsigned char* buffer, unsigned int address);

/* Erases a flash block. Returns TRUE if flash erased successfully,
 * FALSE otherwise
 * Address must be block-aligned.
 */
int flash_erase_block(unsigned int address);

/* Initializes flash */
/* Returns TRUE if flash init OK, FALSE otherwise */
int flash_init(void);

#endif // NANDPRIV_H
